import inspect
import jinja2
from math import floor, log10
import os
import numpy as np
from jinjafy import jinja_env


def jinjafy_template(data,file_in,file_out=None, filters={},template_searchpath=None):
    if template_searchpath:
        file_in = os.path.relpath(file_in, template_searchpath)

    return jinjafy_comment(data, file_in=file_in, file_out=file_out,jinja_tag=None, filters=filters,template_searchpath=template_searchpath)


def jinjafy_comment(data,file_in=None,file_out=None,jinja_tag="jinja",jinja_code=None,trim_whitespace=True,trim_comments=True,comment_char="#",
                    filters={},template_searchpath=None):
    # Extract all comments from the given file and jinjafy them.
    if file_in is None:
        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])
        file_in = module.__file__
    elif not jinja_tag:
        trim_comments=False
        trim_whitespace=False

    if not template_searchpath:
        with open(file_in,'r') as f:
            s = f.read()
            if jinja_tag:
                stag = "<" + jinja_tag + ">"
                etag = "</" + jinja_tag + ">"

                i_start = s.find(stag)
                i_end = s.find(etag)
                s = s[i_start+len(stag):i_end]
            ss = [s]
            if trim_comments:
                ss = [ds.strip()[1:] for ds in s.splitlines() if len(ds.strip()) > 0 and ds.strip()[0] in ["#", "%"] ]
            if trim_whitespace:
                ss = [ds.strip() for ds in ss]

            jinja_code = '\n'.join(ss)

    from jinjafy.snipper import SnipperExtension
    extensions = [SnipperExtension]
    if template_searchpath:
        if not isinstance(template_searchpath, list):
            template_searchpath = [template_searchpath]
        template_searchpath = [ts.replace("\\", "/") for ts in template_searchpath]
        templateLoader = jinja2.FileSystemLoader(searchpath=template_searchpath)
        env = jinja2.Environment(lstrip_blocks=True, trim_blocks=True,loader=templateLoader, extensions=extensions)
    else:
        env = jinja2.Environment(lstrip_blocks=True, trim_blocks=True, extensions=extensions)

    import math
    env.globals['exp'] = math.exp
    env.globals['sqrt'] = math.sqrt
    env.globals['cos'] = math.cos
    env.globals['sin'] = math.sin

    env.globals['mround'] = mround
    env.globals['bold'] = bold
    env.globals['fmat'] = fmat
    env.globals['enumerate'] = enumerate
    env.globals['zip'] = zip
    env.globals['ensure_numpy'] = ensure_numpy
    env.globals['transpose'] = transpose
    import math
    env.globals['ceil'] = math.ceil
    env.globals['floor'] = math.floor


    from pylatexenc import latexencode
    env.globals['utf8tolatex'] = latexencode.utf8tolatex
    env.globals['as_set'] = jinja_env.as_set
    env.globals['as_set_list'] = jinja_env.as_set_list
    env.globals['len'] = jinja_env.mylen
    env.globals['get'] = jinja_env.jget
    env.globals['tolist'] = jinja_env.tolist

    filters['as_set'] =  jinja_env.as_set
    filters['format_list'] =jinja_env.format_list
    filters['format_join'] = jinja_env.format_join
    filters['format_join_enum'] = jinja_env.format_join_enum
    filters['pm'] = lambda x: f" {x}" if x < 0 else f"+{x}"
    filters['bold'] = bold
    filters['capfirst'] = lambda x: (x[0].upper() + x[1:] if len(x) > 1 else x.upper()) if x != None and isinstance(x, str) else x
    filters['lowerfirst'] = lambda x: (x[0].lower() + x[1:] if len(x) > 1 else x.lower()) if x != None and isinstance(x, str) else x
    filters['infty'] = jinja_env.infty
    filters['n2w'] = jinja_env.n2w
    def latex_url(url):
        if not isinstance(url, str):
            return url
        url = url.replace("%", r"\%")
        return url
    filters['latex_url'] = latex_url
    filters['format_list_symbols'] = jinja_env.format_list_symbols
    filters['mround'] = mround
    def eround(val,l):
        x = str(mround(val, l))
        if l == 0:
            return x
        if '.' not in x:
            x = x + "."
        n = l - (len(x) - x.find(".") - 1)
        if n > 0:
            x = x + "0"*n
        return x

    filters['eround'] = eround
    filters['get'] = jinja_env.jget
    filters['flatten'] = jinja_env.flatten
    filters['aan'] = jinja_env.aan
    filters['bracket'] = bracket
    filters['tolist'] = jinja_env.tolist
    filters['rational'] = jinja_env.as_rational
    filters['permute_exam_answers'] = jinja_env.permute_exam_answers
    env.filters.update(filters)

    data['block_start_string'] = '{%'
    if not template_searchpath:
        jinja_out = env.from_string(jinja_code).render(data)
    else:
        file_in = file_in.replace("\\", "/")
        template = env.get_template(file_in)
        jinja_out = template.render(data)

    if file_out is not None:
        with open(file_out,'w',encoding='utf-8') as f:
            # jinja_out = jinja_out.encode('utf-8')

            f.write(jinja_out)
            print("Writing to: " + file_out)

    return jinja_out


def bold(bob,d=True) :
    if not isinstance(bob, str) :
        bob = str(bob)
    if d :
        bob = '\\textbf{' + bob +"}"
    return bob


def fmat(bob,l=2,dobold=False) :
    bob = mround(bob,l)
    bob = bold(bob, dobold)
    return bob

def bracket(s):
    return "{"+str(s)+"}"

def un2str(x, xe, precision=2):
    """pretty print nominal value and uncertainty

        x  - nominal value
        xe - uncertainty
        precision - number of significant digits in uncertainty

        returns shortest string representation of `x +- xe` either as
        x.xx(ee)e+xx
        or as
        xxx.xx(ee)"""
    # base 10 exponents
    x_exp = int(floor(log10(x)))
    xe_exp = int(floor(log10(xe)))

    # uncertainty
    un_exp = xe_exp - precision + 1
    un_int = round(xe * 10 ** (-un_exp))

    # nominal value
    no_exp = un_exp
    no_int = round(x * 10 ** (-no_exp))

    # format - nom(unc)exp
    fieldw = x_exp - no_exp
    fmt = '%%.%df' % fieldw
    result1 = (fmt + '(%.0f)e%d') % (no_int * 10 ** (-fieldw), un_int, x_exp)

    # format - nom(unc)
    fieldw = max(0, -no_exp)
    fmt = '%%.%df' % fieldw
    result2 = (fmt + '(%.0f)') % (no_int * 10 ** no_exp, un_int * 10 ** max(0, un_exp))

    # return shortest representation
    if len(result2) <= len(result1):
        return result2
    else:
        return result1


def mround(val, l=2):
    if not isinstance(l, int):
        return un2str(val, l, 1)
    else:
        if isinstance(val, np.ndarray):
            return np.round(val * 10 ** l) / (10 ** l)
        else:
            return round(val * 10 ** l) / (10 ** l)


def transpose(X):
    return np.transpose( ensure_numpy( X) )


def ensure_numpy(X):
    if type(X) != np.ndarray:
        X = np.asarray(X)
    if X.ndim == 1:
        X = np.transpose( np.expand_dims(X,1) )
    return X


import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import inspect
import os
from datetime import datetime

def subplots(nrows=2, ncols=2, size_inches=(10,12), *args):
    fig,axx = plt.subplots(nrows, ncols, *args)
    fig.set_size_inches(size_inches[0], size_inches[1])
    if not isinstance(axx, list):
        axx = np.asarray(axx, dtype=np.object).reshape((nrows, ncols))
    return fig, axx

def get_colors(palette="dark",max_colors=5):
    return sns.color_palette(palette, max_colors)


def watermark_plot(extra="", nback=2, fz_y=10):
    # from slider.thtools_base import watermark_string
    s = watermark_string(nback=nback)
    plt.figtext(0.05, 0.95, s)



def watermark_string(nback=2):


    tm =  datetime.now().strftime('%b-%d-%I:%M%p')
    # for line in traceback.format_stack():
    #     #     print(line.strip())
    v = inspect.stack()
    ss = []
    j = 0
    for i in range(len(v)):
        if "plot_helpers.py" in v[i].filename: continue
        ss.append( os.path.basename( v[i].filename) )
        j = j + 1
        if j > nback: break
    # from thtools import execute_command
    from jinjafy import execute_command
    v, _ = execute_command("git rev-parse --short HEAD".split())

    ss.append(tm)
    return ('/'.join(ss) + f" @{v}").strip()






from jinjafy import execute_command
import os
from bs4 import BeautifulSoup

def svg2pdf(fin, fout=None, crop=True, text_to_path=False, export_area_page=True):
    """
     -C, --export-area-page                     Area to export is page
       -T, --export-text-to-path                  Convert text to paths (PS/EPS/PDF/SVG)
    """
    # text_to_path = True
    if fout is None:
        fout = fin[:-4] + ".pdf"
    cmd = ['inkscape']
    if export_area_page:
        cmd.append("-C")
    if text_to_path: # Good idea for inkscape which seems to bungle the fonts (space in font names?)
        cmd.append("-T")
    cmd.append(fin)
    cmd.append(f"--export-filename={fout}")
    # '-C', '--without-gui', f'--file={fin}', f'--export-pdf={fout}']
    # cmd = ['inkscape', '-C', '-T', '--without-gui', '--file=%s'%svg_fonts_layers[-1], '--export-pdf=%s' % pdf_nofonts_layers[-1]]
    execute_command(cmd)
    # cmd = f"pdftocairo {fout} -pdf {fout}"
    # execute_command(cmd.split())

    if crop:
        cmd = ['pdfcrop', fout, fout]
        execute_command(cmd)



def pdf2svg(fin, fout, page_no=None):
    '''
    To remove fonts look at
    https://tex.stackexchange.com/questions/23407/how-can-i-convert-text-to-paths-with-pdflatex
    convert to ps and back to pdf
    '''
    if fout is None:
        fout = fin[:-4] + ".svg"

    '''
    pdftocairo -svg C:/Users/tuhe/Documents/02465public/Lectures/Lecture_2/Latex/Lecture_2_NO_SVGS.pdf C:/Users/tuhe/Documents/02465public/Lectures/Lecture_2/Latex/osvgs/tmp/determpath.svg -f 2 -l 2
    
    '''
    cmd = ['pdftocairo', '-svg', fin, fout]
    if page_no is not None:
        if not isinstance(page_no, str):
            page_no = str(page_no)
        cmd += ['-f', str(page_no), '-l', str(page_no)]

    execute_command(cmd)


def pdf2png(fin, fout=None, scale_to=None):
    if fout is None:
        fout = fin[:-4] + ".png"
    fout = fout[:-4]
    cmd = f"pdftocairo -png -singlefile {fin} {fout}"
    if scale_to is not None:
        cmd += f" -scale-to {scale_to}"
    execute_command(cmd.split())
    return fout + ".png"


def pdfcrop(fin, fout=None):
    if fout is None:
        fout = fin
    cmd = f"pdfcrop {fin} {fout}"
    execute_command(cmd.split())



def svg_edit_to_importable(svg_edit_file,verbose=False, keep_background_layer=True):
    assert False
    """
    Take an inkscape file as input and split it into layers.
    CODE NOT IN USE RIGHT NOW; MUST WORK OUT WHAT TO USE IT FOR.
    """
    odir = os.path.dirname(svg_edit_file)
    fn = os.path.basename(svg_edit_file)[:-4]

    pdf_nofonts_base = odir + "/x_do_not_edit_%s-l%s_nofonts.pdf"
    svg_fonts_base =  odir + "/" + SVG_TEXINCLUDE_RELPATH + "/%s-l%s_fonts.svg"
    if not os.path.exists(os.path.dirname(svg_fonts_base)):
        os.mkdir(os.path.dirname(svg_fonts_base))

    pdf_nofonts_layers = []
    svg_fonts_layers = []

    with open(svg_edit_file, 'r', encoding="UTF-8",errors="surrogateescape") as f:
        soup = BeautifulSoup(f, 'xml', from_encoding="UTF-8")
        for i in soup.findAll("image", {'id': 'image4444th'}):
            i.extract()

        layer_labels = []
        for i in soup.findAll("g", {'inkscape:groupmode': 'layer'}):
            if i['inkscape:label'] == "bg_layer":
                #i.extract()
                pass
            else:
                layer_labels.append(i['inkscape:label'])

        for j in range(len(layer_labels)):
            s2 = soup.__copy__()
            for i in s2.findAll("g", {'inkscape:groupmode': 'layer'}):
                if layer_labels[j] == i['inkscape:label'] or i['inkscape:label'] == "bg_layer":
                    pass
                else:
                    i.extract()
            # now you got an image only with this layer. save it.
            layer_number = layer_labels[j].split(" ").pop()
            pdf_nofonts_layers.append(pdf_nofonts_base%(fn,layer_number))
            svg_fonts_layers.append(svg_fonts_base % (fn, layer_number))

            with open(svg_fonts_layers[-1], 'bw') as f2:
                f2.write(s2.encode("UTF-8"))

            cmd = ['inkscape', '-C', '-T', '--without-gui', '--file=%s'%svg_fonts_layers[-1], '--export-pdf=%s' % pdf_nofonts_layers[-1]]
            execute_command(cmd)

    if verbose:
        print("svg_edit_to_importable called. Converting svg file\n  > %s\nto files:"%svg_edit_file)
        for s in pdf_nofonts_layers + svg_fonts_layers:
            print("  > " + s)

    return pdf_nofonts_layers, svg_fonts_layers



#!python
# No, do this instead: https://setuptools.readthedocs.io/en/latest/userguide/entry_point.html
# The above makes the script executable.

import clize
import os
import sys
import shutil

base_slide = """ 
\\documentclass[aspectratio=43]{beamer}
\\usepackage{etoolbox}
\\newtoggle{overlabel_includesvgs}
\\newtoggle{overlabel_includelabels}
\\toggletrue{overlabel_includesvgs}
\\toggletrue{overlabel_includelabels}
\\input{beamer_slider_preamble.tex}

\\title{Example slide show}
\\author{Tue Herlau}
\\begin{document}
\\begin{frame}
\\maketitle
\\end{frame}

\\begin{frame}\\osvg{myoverlay} % Use the \\osvg{labelname} - tag to create new overlays. Run slider and check the ./osvgs directory for the svg files!
\\title{Slide with an overlay}
This is some example text!
\\end{frame}

\\end{document}
"""

def slider_init(latexfile=None):
    # return
    # print("Initializing da slides.")
    wdir = os.getcwd()
    print(wdir)
    if latexfile == None:
        latexfile = "index.tex"
    if not latexfile.endswith(".tex"):
        latexfile += ".tex"
    latexfile = os.path.join(wdir, latexfile)
    if os.path.exists(latexfile):
        print("File already exists", latexfile)
        # sys.exit()
    # Done with the introductory bullshit.

    if not os.path.isdir(os.path.dirname(latexfile)):
        os.makedirs(os.path.dirname(latexfile))

    import jinja2
    with open(latexfile, 'w') as f:
        f.write(base_slide)

    print("Initializing with", latexfile)

    # jinja2.Environment().from_string(base_slide)
    from slider.slide import set_svg_background_images
    set_svg_background_images(latexfile, clean_temporary_files=True)

if __name__ == "__main__":
    # slider_init("../../test/index.tex")
    # from slider.latexutils import latexmk
    # import slider
    clize.run(slider_init)



import os
import pkg_resources  # part of setuptools
from jinjafy import execute_command
import subprocess
from datetime import datetime

bibliography = """
@online{%s,
	title={%s},
	url={%s},
	urldate = {%s}, 
	month={9},
	publisher={Technical University of Denmark (DTU)},
	author={Tue Herlau},
	year={%s},
}
"""
# 	comments={See url{asdfsdaf} for examples},

def make_bibliography(setup_py, outfile=None):
    if not os.path.isfile(setup_py) or os.path.basename(setup_py) != "setup.py":
        raise Exception("No setup.py")

    v = subprocess.check_output(f"cd {os.path.dirname(setup_py)} && python setup.py --name --version --url --description", shell=True).decode("utf-8").splitlines()
    v = [s.strip() for s in v]
    name, version, url, description = v

    ex = "\\texttt{pip install " + name + "}"
    title = f"{name.capitalize()} ({version}): {ex}"
    date = datetime.today().strftime('%Y-%m-%d')
    year = datetime.today().strftime('%Y')

    key = name.replace("-", "_")
    s = (bibliography%(key, title, url, date, year) ).strip()
    if outfile is not None:
        if os.path.isdir(outfile):
            outfile = os.path.join(outfile, name.replace("-",'_') + ".bib")
        with open(outfile, 'w') as f:
            f.write(s)
        print(f"[Writing bibliography with entry '{key}' to]", outfile)

    return s



import  clize
import os
import sys
import shutil
from slider.slide import set_svg_background_images
import click

def confirm_start_new_project(latexfile, force=False):
    try:
        if force or click.confirm(f"Do you want to create a new Slider LaTeX file named {latexfile}?", abort=True):
            # print("Starting new project")
            from slider.slider_init import slider_init
            slider_init(latexfile)

    except click.exceptions.Abort as e:
        sys.exit()


def slider_cli(latexfile=None, force=False, verbose=False):
    """
    Slider software for manipulating overlay-svg images.
    To get started, first start a slider project by creating a new folder and running

    > python -m slider index.tex

    This will create a bunch of files including a folder named osvgs. This is where you keep the slides!

    When you edit/change overlays, remember to run

    > python -m slider index.tex
    > python -m slider

    to keep everything synchronized.
    You can add new overlays by simply using the LaTeX \osvg{labelname}-tag on new slides (and running slider)
    Edit the overlays by looking in the \osvg-folder, in this case osvg/labelname.svg.

    Remember the overlays by default import the content of the slides (useful if you want to move existing equations around)
    so remember to remove non-wanted contents.
    When done, run slider again to keep everything in sync.

    :param latexfile:
    :param force:
    :param verbose:
    """

    # print("Initializing da slides.")
    wdir = os.getcwd()
    print(wdir)
    if latexfile == None:
        print("Trying to manually detect main latex file.")
        import glob
        files = glob.glob("*.tex")
        mfiles = []
        for name in files:
            with open(name, 'r') as f:
                lines = [l.strip() for l in f.read().splitlines()]
            s = "\n".join([l for l in lines if not l.startswith("%")] )
            if "\\begin{document}" in s and "{beamer}" in s and "_NO_SVGS" not in name:
                print("Main file found!")
                mfiles.append(name)
        if len(mfiles) != 1:
            print("Many candidate files found")
            print(mfiles)
            sys.exit()
        else:
            latexfile = mfiles[0]
        # latexfile = "index.tex"
    if not latexfile.endswith(".tex"):
        latexfile += ".tex"
    latexfile = os.path.join(wdir, latexfile)
    if os.path.exists(latexfile):
        # print("File already exists:", latexfile)
        # print("Doing the slide-stuff.")
        set_svg_background_images(lecture_tex=latexfile)
    else:
        confirm_start_new_project(latexfile=latexfile, force=force)


def clize_main_entry_point():
    """
    I collect this in one function to make a single entry point regardless of where
    > slider
    or
    > python -m slider

    is used.

    :return:
    """
    clize.run(slider_cli)


if __name__ == '__main__':
    clize_main_entry_point()



import os
import glob


def beamer_nup(pdf_file, nup=6, output=None, a4paper=True):
    # pdf_file = pdf_file[:-4] + ".pdf"
    if nup not in [1, 2, 3, 4, 6]:
        assert False

    jinja = os.path.dirname( __file__ ) + "/../../jinja"
    if os.path.isdir(jinja):
        js = {}
        for name in glob.glob(jinja + "/*.tex"):
            with open(name, 'r') as f:
                js[os.path.basename(name)[:-4] ] = f.read()

        s = ""
        for k, v in js.items():
            v = v.replace("\\", "\\\\")

            s += f'{k} = """\n' + v + '\n"""' + "\n"

        with open(os.path.dirname(__file__) + "/jinjastrings/generated.py", 'w') as f:
            s = "# WARNING! THIS FILE IS AUTOMATICALLY GENERATED! ALL CHANGES WILL BE WIPED. SEE JINJA DIRECTORY\n"*10  + s
            f.write(s)



    from slider.jinjastrings.generated import lecture_collector_partial

    import jinja2

    import tempfile
    # tempfile.gettempdir()
    tmp = tempfile.TemporaryDirectory().name
    os.mkdir(tmp)
    import shutil
    dest_pdf = tmp + "/" + os.path.basename(pdf_file)
    shutil.copyfile(pdf_file, dest_pdf)

    import jinja2
    data = {'a4': False,
            'twoup': False,
            'sixup': False,
            'a4paper': a4paper}
    if nup == 1:
        data['a4'] = True
    if nup == 2:
        data['twoup'] = True
    if nup == 3:
        data['threeup'] = True
    if nup == 4:
        data['fourup'] = True
    if nup == 6:
        data['sixup'] = True
    data['frame'] = True

    data['pdffiles'] = [os.path.basename(dest_pdf)]

    # data = {'hello': 'world'}
    print(tmp)
    s = jinja2.Environment().from_string(lecture_collector_partial).render(data)

    with open(tmp +"/nup.tex", 'w') as f:
        f.write(s)

    from slider import latexmk
    latexmk(tmp +"/nup.tex", shell=True)

    if output == None:
        output = os.path.dirname(pdf_file) + "/" + os.path.basename(pdf_file)[:-4] + f"_{nup}up.pdf"
    shutil.move(tmp +"/nup.pdf", output)
    print(tmp)
    print("[Beamer-nup] Wrote output to", output)
    return output



if __name__ == "__main__":
    beamer_nup("../../examples/new_project/index.pdf")




from jinjafy.jinjafy import jinjafy_comment
from jinjafy.jinjafy import jinjafy_template
from jinjafy.jinja_matlab_load import matlab_load
# from slider import latexmk
from jinjafy.textools import mat2table
import subprocess
# import os
import platform
# from subprocess import subprocess


def execute_command(command, shell=True):
    if not isinstance(command, list):
        command = [command]
    if not platform.uname()[0] == "Linux":
        result = subprocess.run(command, stdout=subprocess.PIPE, shell=shell)
        out = result.stdout
    else:
        out = subprocess.check_output(command, shell=shell)
    s = out.decode("utf-8")
    OK = True
    return s, OK


# def get_system_name():
#     if is_win():
#         return "Win"
#     if is_compute():
#         return "thinlinc.compute.dtu.dk"
#     if is_cogsys_cluster():
#         return "cogys cluster"

# def execute_command(command, shell=True):
#     if not isinstance(command, list):
#         command = [command]
#     # if not is_compute():
#     # result = subprocess.run(command, stdout=subprocess.PIPE, shell=shell)
#     # out = result.stdout
#     # else:
#     out = subprocess.check_output(command, shell=shell)
#     s = out.decode("utf-8")
#     OK = True
#     return s, OK


import numpy as np
import scipy.io as spio

def matlab_load(mfile):
    j = mfile.rfind('.')
    if j > -1:
        ex = mfile[j + 1:]
        base = mfile[:j]
    else:
        ex = ''
        base = mfile
    mat = loadmat(base + '.mat')
    mat = uuroll(mat)
    mat = fix_1_arrays(mat)
    mat = fix_strings(mat)
    mat = fix_simple_lists(mat)
    return mat


def loadmat(filename):
    '''
    this function should be called instead of direct spio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects
    '''
    data = spio.loadmat(filename,struct_as_record=False)
    data2 = _check_keys(data)
    return data2


def _check_keys(dd):
    '''
    checks if entries in dictionary are mat-objects. If yes
    todict is called to change them to nested dictionaries
    '''
    if isinstance(dd, spio.matlab.mio5_params.mat_struct):
        dd = _check_keys(_todict(dd))
    elif type(dd) == dict:
        for key in dd:
            kv = flist(dd[key])
            if type( kv ) == spio.matlab.mio5_params.mat_struct:
                dd[key] = _check_keys(kv)
            else:
                dd[key] = _check_keys(dd[key])
    elif type(dd) == list:
        dd = [_check_keys(l) for l in dd]
    elif type(dd) == np.ndarray:
        if dd.dtype.str == '|O' and dd.size > 0:
            if type( flist(dd.flat[0]) ) == spio.matlab.mio5_params.mat_struct:
                for i in range( dd.size ):
                    dd.flat[i] = _check_keys( flist( dd.flat[i]) )
        else:
            for i in range(dd.size):
                dd.flat[i] = _check_keys(dd.flat[i])

    return dd

def fix_simple_lists(l):
    if type(l) == dict:
        for k,v in l.items():
            l[k] = fix_simple_lists(v)
    elif type(l) == np.ndarray and l.dtype.name == "uint8" and l.shape[0] == 1 and l.ndim == 2:
        # l = l.tolist()
        l = l.tolist()[0]
    return l

def apply_recursively(l, myfun):
    if type(l) == dict:
        for k,v in l.items():
            l[k] = apply_recursively(v, myfun)
    elif type(l) == np.ndarray and l.dtype.str == '|O' and l.size > 0:
        for i in range( l.size ):
            l.flat[i] = apply_recursively( l.flat[i], myfun)
    else:
        l = myfun(l)
    return l


def fix_1_arrays(l):
    def _fix_1_arrays(l):
        if type(l) == np.ndarray and l.size == 1 and np.issubdtype(l.dtype, np.number):
            l = l.flat[0]
        return l
    l = apply_recursively(l, _fix_1_arrays)
    return l


def fix_strings(l):
    if type(l) == dict:
        for k,v in l.items():
            l[k] = fix_strings(v)
    elif type(l) == np.ndarray and l.size > 0:
        tp = type(superpop(l.flat[0]))
        if tp == str or tp == np.str_:
            l = [superpop(x) for x in l.flat ]
            if len(l) == 1:
                l = l.pop()
    return l


def superpop(l):
    if type(l) == list and len(l) == 1:
        return superpop(l[0])
    if type(l) == np.ndarray and l.size == 1:
        return superpop(l.tolist())
    return l


def flist(l):
    if type(l) == list and len(l) == 1:
        l = flist( l.pop() )

    if type(l) == np.ndarray and l.dtype.name == "object":
        l3 = [flist(v) for v in l.flat]
        l = flist( l3 )
    return l


def _todict(matobj):
    '''
    A recursive function which constructs from matobjects nested dictionaries
    '''
    dict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, spio.matlab.mio5_params.mat_struct):
            dict[strg] = _todict(elem)
        else:
            dict[strg] = elem
    return dict


def uuroll(v):
    if type(v) is dict:
        for key,val in v.items():
            v[key] = uuroll(val)
    if type(v) is np.ndarray or type(v) is np.array:
        for j in range(v.size):
            v.flat[j] = uuroll(v.flat[j])
    return v


def uroll(mat):
    for k in mat.keys():
        v = mat[k]
        v = uuroll(v)
        mat[k] = v
    return mat


# Control import of slides from pdf to svg-editable format.
# inkscape -z -f "Input.pdf" -l "Output.svg"
# https://github.com/eea/odfpy
import os
import shutil
from jinjafy import jinjafy_comment
from bs4 import BeautifulSoup
import glob
from jinjafy import execute_command

CDIR = os.path.dirname(os.path.realpath(__file__))
CDIR = CDIR.replace('\\','/')

SVG_EDIT_RELPATH = "osvgs" # files that are supposed to be edited goes here.
SVG_TMP_RELPATH = "tmp" # various files that can be flat out deleted goes here
SVG_TEXINCLUDE_RELPATH = "do_not_edit" # the no_fonts version and the pure (+fonts) versions goes here
DTU_beamer_base = CDIR +"/DTU_Beamer_files"
BLANK_PNG =DTU_beamer_base + "/blank.png"

def ensure_dir(dname):
    assert False
    if not os.path.exists(dname):
        os.mkdir(dname)

def join_pdfs(slide_deck_pdf, outfile):
    assert False
    dn = os.path.dirname(slide_deck_pdf[0])
    files = [os.path.relpath(os.path.dirname(pdf), start=dn) + "/" + os.path.basename(pdf) for pdf in slide_deck_pdf]
    outf = os.path.relpath(os.path.dirname(outfile), start=dn) + "/" + os.path.basename(outfile)
    cmd = "cd " + dn + " && pdftk " + " ".join(files) + " cat output " + outf
    execute_command(cmd.split())


def li_import(slide_deck_pdf, tex_output_path=None, num_to_take=None, force=False, svg_pfix="osvg", svg_height=743.75, svg_width=992.5,
              svg_converted_slides="svg_converted_slides.tex"):
    assert False
    '''
    svg_height and svg_width are used to scale the converted image. This is useful because otherwise the viewbox
    will fail to match the DTU template. I.e. these numbers will generally change dependent on the LaTeX template.

    :param slide_deck_pdf:
    :param tex_output_path:
    :param num_to_take:
    :param force:
    :param svg_pfix:
    :param svg_height:
    :param svg_width:
    :return:
    '''
    # take this slide deck. Generate beamer, svg output.

    if isinstance(slide_deck_pdf, list):
        dn = os.path.dirname(slide_deck_pdf[0])
        ofile = os.path.join(dn, "tmp.pdf")
        join_pdfs(slide_deck_pdf, ofile)
        slide_deck_pdf = ofile

    if tex_output_path is None:
        tex_output_path = slide_deck_pdf[:-4]+"_output.tex"

    output_dir = os.path.dirname(tex_output_path)

    # if output_dir is None:
    #     output_dir = os.path.dirname(slide_deck_pdf)
    assert(os.path.exists(output_dir))
    svg_tmp_dir = output_dir + "/" + SVG_EDIT_RELPATH +"/" + SVG_TMP_RELPATH
    svg_texinclude_dir = output_dir + "/" + SVG_EDIT_RELPATH +"/" +SVG_TEXINCLUDE_RELPATH
    svg_edit_dir = output_dir + "/" +SVG_EDIT_RELPATH
    tex_output_file = os.path.basename(tex_output_path)

    print("Converting slides to output: " + tex_output_file)
    if os.path.exists(output_dir + "/" + tex_output_file) or glob.glob(svg_edit_dir + "/*.svg"):
        print("Non-empty output directory...")
        if not force:
            raise Exception("Non-empty output directory. Please clean")

    ensure_dir(svg_edit_dir)
    ensure_dir(svg_tmp_dir)
    ensure_dir(svg_texinclude_dir)

    if not output_dir:
        raise Exception("Must specify output directory!")

    lecture_tex_out = move_template_files(output_dir, tex_output_file)
    osvgs_basename = []
    print("Splitting slide deck into images...")

    # slide_deck_split_svg = slidedeck_to_images(slide_deck_pdf, svg_tmp_dir + "/" + svg_pfix + "-%i.svg",
    #                                            num_to_take=num_to_take)
    slide_deck_split_svg = slidedeck_to_images(slide_deck_pdf, svg_tmp_dir+"/"+svg_pfix+"-%i.svg", num_to_take=num_to_take)
    print("Converting svg to osvg..")
    for i,osvg in enumerate(slide_deck_split_svg):
        dosvg = raw_svg_to_osvg(osvg, overwrite_existing=True, height=svg_height, width=svg_width)
        osvgs_basename.append(dosvg)

    print("jinjafying and cleaning...")
    # osvgs_basename = osvgs_basename[3:] # Drop first 3 slides; automatically generated.
    data = {'osvgs_basename' : osvgs_basename}
    s = jinjafy_comment(data, jinja_tag="jinja1")
    with open(output_dir + "/%s"%svg_converted_slides, 'w') as f:
        f.write(s)
    return lecture_tex_out

# <jinja1>
# {% for sf in osvgs_basename %}
# \begin{frame}\osvg{{"{"}}{{sf}}{{"}"}}
# % add content here
# \end{frame}
# {% endfor %}
# </jinja1>
# \begin{textblock}{1}(0,0)
# 	\includesvg[width=1.0\linewidth]{{"{"}}{{sf}}{{"}"}}
# \end{textblock}\overlabel{ {{sf}} }
# SVG editable file (i.e. with background image) to file which can be imported into
# the .tex file.

'''
Take a raw svg in the tmp directory and compile it into the nice svg format with empty, white background.
This can be used when importing a new slide deck or when inserting a new overlabel tag somewhere in a
tex document.
'''
def raw_svg_to_osvg(raw_svg_file, overwrite_existing=False, height=None, width=None):
    svg_tmp_dir = os.path.dirname(raw_svg_file)
    svg_edit_dir = os.path.dirname(svg_tmp_dir)
    ofile_edit = svg_edit_dir + "/" + os.path.basename(raw_svg_file)
    # if height is not None and width is not None:
    #     svg_set_hw_(raw_svg_file, raw_svg_file,height=height, width=width)

    ofile_fonts_pure = rm_svg_bg(svg_input=raw_svg_file, svg_output=ofile_edit, height=height, width=width)
    png_file = svg_tmp_dir + "/" + os.path.basename(ofile_fonts_pure)[:-4] + ".png"
    shutil.copyfile(BLANK_PNG, png_file)
    ofile_edit = add_png_background_to_svg(svg_input=ofile_fonts_pure, svg_output=None, png_file=png_file)
    pdf_nofonts, svg_fonts = svg_edit_to_importable(ofile_edit)
    osvgs_basename = os.path.basename(raw_svg_file)[:-4]
    return osvgs_basename


'''
Related to li_import. 
Set the width/height of an imported slide svg image in case it does not match the DTU template. 
'''
def svg_set_hw_(svg_in, svg_out, height, width):
    assert False
    print(f"HW fix [{height} {width}] > {svg_in} -> {svg_out}")

    with open(svg_in, 'r', encoding="UTF-8") as f:
        soup = BeautifulSoup(f, 'xml', from_encoding="UTF-8")
        tags = soup.find_all("svg")
        assert (len(tags) == 1)
        tag = tags[0]
        tag['height'] = str(height)
        tag['width'] = str(width)
        tag['viewBox'] = f"0 0 {height} {width}"
    # print([svg_input, logo_rem, bg_rem, tx_rem])
    with open(svg_out, 'bw') as f:
        f.write(soup.encode("UTF-8"))
    return
    with open(svg_in, 'r', encoding="UTF-8", errors="surrogateescape") as f:
        soup = BeautifulSoup(f, 'xml', from_encoding="UTF-8")
        tags = soup.find_all("svg")
        assert(len(tags) == 1)
        tag = tags[0]
        tag['height'] = str(height)
        tag['width'] = str(width)
        tag['viewBox'] = f"0 0 {height} {width}"
        s2 = soup.__copy__()
        # sout = s2.encode("UTF-8")
    # f.close()
    with open(svg_out, 'w', encoding="UTF-8") as f2:
        f2.write(str(s2))

def svg_check_background_layer(svg_edit_file, verbose=False):
    assert False

    # Check if svg background layer is pointing to the right .png file.
    # this may not be the case sometimes because svg files are moved, etc. which overwrite the default
    # background .png path.
    with open(svg_edit_file, 'r', encoding="UTF-8",errors="surrogateescape") as f:
        soup = BeautifulSoup(f, 'xml', from_encoding="UTF-8")
        g = None
        for i in soup.findAll("g", {'inkscape:groupmode': 'layer'}):
            if i['inkscape:label'] == "bg_layer":
                g = i
                break
        ok = True
        bgim = g.find("image")
        bg_png = bgim['xlink:href']
        real_png = os.path.dirname(svg_edit_file) + "/" + SVG_TMP_RELPATH + "/" + os.path.basename(svg_edit_file)[:-4] + ".png"
        real_png = os.path.relpath(real_png, start=os.path.dirname( svg_edit_file) )
        bg_png = os.path.relpath(bg_png,start=os.path.dirname( svg_edit_file)  )
        if real_png != bg_png:
            print("slider:warning> Bungled background png image in " + svg_edit_file)
            s = jinjafy_comment({'png_file': real_png}, jinja_tag="jinja3")
            new_img = BeautifulSoup(s, "html.parser")
            g.insert_after( new_img)
            g.unwrap()
            bgim.unwrap()

            with open(svg_edit_file[:-4]+"_test.svg", "w") as f2:
                f2.write(soup.prettify(formatter="xml"))

# <jinja3>
# <g inkscape:groupmode="layer" id="layer1" inkscape:label="bg_layer" style="display:inline" sodipodi:insensitive="true">
#      <image
#        xlink:href="{{png_file}}"
#        width="100%"
#        height="100%"
#        preserveAspectRatio="none"
#        style="image-rendering:optimizeQuality"
#        id="image4444th"
#        x="0"
#        y="0" />
#  </g>
# </jinja3>



def svg_edit_to_importable(svg_edit_file,verbose=False):
    odir = os.path.dirname(svg_edit_file)
    fn = os.path.basename(svg_edit_file)[:-4]

    pdf_nofonts_base = odir + "/x_do_not_edit_%s-l%s_nofonts.pdf"
    svg_fonts_base =  odir + "/" + SVG_TEXINCLUDE_RELPATH + "/%s-l%s_fonts.svg"
    if not os.path.exists(os.path.dirname(svg_fonts_base)):
        os.mkdir(os.path.dirname(svg_fonts_base))

    pdf_nofonts_layers = []
    svg_fonts_layers = []

    with open(svg_edit_file, 'r', encoding="UTF-8",errors="surrogateescape") as f:
        soup = BeautifulSoup(f, 'xml', from_encoding="UTF-8")

        for i in soup.findAll("image", {'id': 'image4444th'}):
            i.extract()

        layer_labels = []
        for i in soup.findAll("g", {'inkscape:groupmode': 'layer'}):
            if i['inkscape:label'] == "bg_layer":
                #i.extract()
                pass
            else:
                layer_labels.append(i['inkscape:label'])

        for j in range(len(layer_labels)):
            s2 = soup.__copy__()
            for i in s2.findAll("g", {'inkscape:groupmode': 'layer'}):
                if layer_labels[j] == i['inkscape:label'] or i['inkscape:label'] == "bg_layer":
                    pass
                else:
                    i.extract()
            # now you got an image only with this layer. save it.
            layer_number = layer_labels[j].split(" ").pop()
            pdf_nofonts_layers.append(pdf_nofonts_base%(fn,layer_number))
            svg_fonts_layers.append(svg_fonts_base % (fn, layer_number))

            with open(svg_fonts_layers[-1], 'bw') as f2:
                f2.write(s2.encode("UTF-8"))

            from slider.convert import svg2pdf
            svg2pdf(svg_fonts_layers[-1], fout=pdf_nofonts_layers[-1], crop=False, text_to_path=True, export_area_page=True)
            # cmd = ['inkscape', '-C', '-T', '--without-gui', '--file=%s'%svg_fonts_layers[-1], '--export-pdf=%s' % pdf_nofonts_layers[-1]]

    if verbose:
        print("svg_edit_to_importable called. converted svg file\n  > %s\nto files:"%svg_edit_file)
        for s in pdf_nofonts_layers + svg_fonts_layers:
            print("  > " + s)

    return pdf_nofonts_layers, svg_fonts_layers

# <jinja2>
# {{svg_start}}
# <g inkscape:groupmode="layer" id="layer1" inkscape:label="bg_layer" style="display:inline" sodipodi:insensitive="true">
#      <image
#        xlink:href="{{png_file}}"
#        width="100%"
#        height="100%"
#        preserveAspectRatio="none"
#        style="image-rendering:optimizeQuality"
#        id="image4444th"
#        x="0"
#        y="0" />
#  </g>
# <g inkscape:groupmode="layer"
#     id="layer2"
#     inkscape:label="Layer 1"
#     style="display:inline">
# {{svg_end}}
# </g></svg>
# </jinja2>
def add_png_background_to_svg(svg_input, png_file, svg_output=None):
    if not svg_output: svg_output = svg_input
    rp = os.path.relpath(png_file, os.path.commonprefix([svg_output, png_file]))
    rp = rp.replace("\\", "/")

    with open(svg_input,'r', encoding="UTF-8") as f:
        svg = f.read()
    mds = "</metadata>"
    mds_id = svg.find(mds)
    if mds_id < 0:
        # file has no meta data.
        j = svg.find(">", svg.find("<svg"))
        svg = svg[:j+1] + "<metadata></metadata>" +svg[j+1:]
        mds_id = svg.find(mds)

    mds_dex = mds_id+len(mds)

    data = {'png_file' : rp, 'svg_start': svg[:mds_dex],'svg_end': svg[mds_dex:]}
    svg = jinjafy_comment(data, jinja_tag="jinja2")

    si = svg.find("<svg") + 4
    dsvg = ' xmlns:sodipodi = "http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"\n xmlns:inkscape = "http://www.inkscape.org/namespaces/inkscape" \n'
    svg = svg[:si] + dsvg + svg[si:]

    with open(svg_output, 'w',encoding="UTF-8") as f:
        f.write(svg)
    return svg_output


def slidedeck_to_images(slide_deck_pdf, base_out_pattern, num_to_take=None):
    assert False
    if not os.path.exists(os.path.dirname(base_out_pattern)):
        os.mkdir(os.path.dirname(base_out_pattern))
    num_pages = num_pages_in_pdf(slide_deck_pdf)
    opat = base_out_pattern[:-4] + "_tmp.pdf"
    opat.replace("%i", '%d')
    cmd = f"pdftk {slide_deck_pdf} burst output {opat} compress"
    print("pdftk splitting into ", num_pages)

    execute_command(cmd.split())
    outfiles = []
    slide_deck_split_pdf = [base_out_pattern[:-4] % (i + 1) + "_tmp.pdf" for i in range(num_pages)]
    for i, opdf in enumerate(slide_deck_split_pdf):
        print("convertion", i, opdf)
        ofile = base_out_pattern %(i+1)
        cmd = f"pdf2svg {opdf} {ofile}"
        execute_command(cmd.split())
        b = os.path.getsize(ofile)
        # print(b)
        if b == 0:
            print("Skipping this file because it has size 0...")
        else:
            outfiles.append(ofile)

    return outfiles

def slidedeck_to_images_DEFUNCT(slide_deck_pdf, base_out_pattern, num_to_take=None):
    assert False
    if not os.path.exists(os.path.dirname(base_out_pattern)):
        os.mkdir(os.path.dirname(base_out_pattern))

    num_pages = num_pages_in_pdf(slide_deck_pdf)
    slide_deck_split_pdf = [base_out_pattern[:-4] % (i + 1) + "_tmp.pdf" for i in range(num_pages)]
    if num_to_take: slide_deck_split_pdf = slide_deck_split_pdf[0:num_to_take]

    outfiles = []
    for i, opdf in enumerate(slide_deck_split_pdf):
        print("convertion", i, opdf)
        ofile = base_out_pattern %(i+1)
        slide_to_image(slide_deck_pdf, ofile, page_to_take=i+1)
        outfiles.append(ofile)
    return outfiles

def num_pages_in_pdf(pdf_file):
    assert False
    cmd = ['pdftk', '%s' % pdf_file, 'dump_data']
    ss = execute_command(cmd)[0].splitlines()
    s = int([s for s in ss if 'NumberOfPages' in s].pop().split()[-1])
    return s


def slide_to_image(slide_deck_pdf, output, page_to_take=1, use_inkscape=True):
    if not os.path.exists(os.path.dirname(output)):
        os.mkdir(os.path.dirname(output))
    slide_deck_split_pdf = output[:-4] + "_tmp.pdf"
    ext = output[-3:]
    if ext == "svg":
        from slider.convert import pdf2svg
        pdf2svg(slide_deck_pdf, fout=output, page_no=page_to_take)
        # cmd = ['pdftk', '%s' % slide_deck_pdf, 'cat', '%i' % page_to_take, 'output', '%s' % slide_deck_split_pdf]
        # # page_to_take = 1

        # if use_inkscape:
        #     cmd = ['inkscape', '-C', '--without-gui', '--file=%s' % slide_deck_split_pdf, '-l', '%s' % output]
        # else:
        #     cmd = ['pdf2svg', slide_deck_split_pdf, output]
    else:
        if os.path.exists(output):
            os.remove(output)
        cmd = ("pdftocairo -png -f %i -l %i"% (page_to_take, page_to_take)).split(" ") + [slide_deck_pdf, output]
        execute_command(cmd)
    if ext == "png":
        png_with_postfix = glob.glob(output + "-*.png")
        if not png_with_postfix:
            print("WARNING! no png generated.")
            print(output)
        else:
            png_with_postfix = png_with_postfix.pop()
            shutil.move(png_with_postfix, output)

    return output


def move_template_files(output_dir="examples/output", output_tex_file=None):
    files_to_move = ["tex_dtu_logo.pdf", "tex_dtu_compute_a_uk.pdf", "tex_dtu_frise.pdf", "dtucolours.tex",
                     "beamerthemeDTU.sty", "beamerfontthemeDTU.sty","beamercolorthemeDTU.sty",
                     "beamerinnerthemeDTU.sty", "beamerouterthemeDTU.sty", "departments.tex", "tex_compute_uk.pdf",
                     "02450_beamer_preamble.tex",  # Deprecated.
                     'beamer_slider_preamble.tex', # The current version.
                     ]
    sd = list( zip(files_to_move, files_to_move) )
    if output_tex_file:
        sd.append( ("02450_lectures_base.tex", output_tex_file))
    for (source,dest) in sd:
        shutil.copy(DTU_beamer_base + "/" + source, output_dir + "/" + dest)

    if output_tex_file:
        lecture_tex_out = output_dir + "/" + output_tex_file
    else:
        lecture_tex_out = None

    return lecture_tex_out


def rm_svg_bg(svg_input, svg_output=None, fix_bg=True, fix_txt=True, fix_logo=True, height=None, width=None):
    logo_rem = 0
    tx_rem = 0
    bg_rem = 0
    if not svg_output:
        svg_output = svg_input

    with open(svg_input, 'r', encoding="UTF-8") as f:
        soup = BeautifulSoup(f, 'xml', from_encoding="UTF-8")
        BG_white = ["fill:#ffffff;fill-opacity:1;fill-rule:nonzero;stroke:none", "fill:#ffffff;fill-opacity:1;fill-rule:evenodd;stroke:none"]
        for bgw in BG_white:
            gg = soup.findAll("path", {"style" : bgw})
            for g in gg:
                if not fix_bg: break
                g['style'] = bgw.replace("opacity:1", "opacity:0")
                bg_rem += 1
                if bg_rem >= 2: break

        dtulogo = soup.findAll("image")
        for i in dtulogo:
            if "iVBORw0KGgoAAAANSUhEUgAABawAAAFcCAYAAAAkg" in i['xlink:href'] and fix_logo:
                i.extract()
                logo_rem += 1

        btx = ["font-variant:normal;font-weight:bold;font-size:8px;font-family:Verdana;-inkscape-font-specification:Verdana-Bold;writing-mode:lr-tb;fill:#000000;fill-opacity:1;fill-rule:nonzero;stroke:none",
               "font-variant:normal;font-weight:normal;font-size:9px;font-family:Verdana;-inkscape-font-specification:Verdana;writing-mode:lr-tb;fill:#000000;fill-opacity:1;fill-rule:nonzero;stroke:none",
               "font-variant:normal;font-weight:bold;font-size:9px;font-family:Verdana;-inkscape-font-specification:Verdana-Bold;writing-mode:lr-tb;fill:#000000;fill-opacity:1;fill-rule:nonzero;stroke:none",
               "font-variant:normal;font-weight:bold;font-size:8px;font-family:Arial;-inkscape-font-specification:Arial-BoldMT;writing-mode:lr-tb;fill:#000000;fill-opacity:1;fill-rule:nonzero;stroke:none",
               "font-variant:normal;font-weight:normal;font-size:9px;font-family:Arial;-inkscape-font-specification:ArialMT;writing-mode:lr-tb;fill:#000000;fill-opacity:1;fill-rule:nonzero;stroke:none"]

        for j,style in enumerate(btx):
            if not fix_txt:
                break
            for tx in soup.findAll("text", {"style": style}):
                tx.extract()
                tx_rem += 1

        # soup = BeautifulSoup(f, 'xml', from_encoding="UTF-8")
        if height is not None and width is not None:
            ''' 
            We are doing this if the svg is being imported and the height/width might not match the DTU template viewbox. 
                        
            '''
            tags = soup.find_all("svg")
            if len(tags) != 1:
                a = 1234

            assert (len(tags) == 1)
            tag = tags[0]
            tag['height'] = str(height)
            tag['width'] = str(width)
            tag['viewBox'] = f"0 0 {height} {width}"


    print([svg_input, logo_rem, bg_rem, tx_rem])
    with open(svg_output, 'bw') as f:
        f.write(soup.encode("UTF-8"))
    return svg_output


if __name__ == "__main__":
    print("operating...")
    lecture_tex_out = li_import("examples/ex1/Lecture11.pdf", output_dir="examples/output", num_to_take=3)
    print("Wrote new main file: " + lecture_tex_out)